;======================================================================
; NUL2 version 1.0
; Published in The DOS Authority
; Copyright (c) 1992, Robert L. Hummel
;----------------------------------------------------------------------
; This program demostrates device driver structure and operation. It
; is configured as a character device and may be accessed from DOS as
; either "NUL2" or "NUL2:". It behaves like the NULL device in that it
; accepts all input and generates only EOFs as output. In addition, it
; displays the calls it receives on the top line of the current
; display.
;
; The executable should be created using the following commands:
;   MASM NUL2;
;   LINK NUL2;
;   EXE2BIN NUL2.EXE NUL2.SYS
;
; To use, load in your CONFIG.SYS file with the line:
;   DEVICE=[d:\path\]NUL2.SYS
;----------------------------------------------------------------------

;----------------------------------------------------------------------
; This macro creates a counted string of the type n,"...str..." where n
; is the length of the string in bytes.
;----------------------------------------------------------------------
LENSTR		MACRO	STRING
		LOCAL	LABEL1

		DB	OFFSET LABEL1 - $ - 1
		DB	STRING
LABEL1		EQU	$

		ENDM

;======================================================================
; Configured in COM (memory image) format.
;----------------------------------------------------------------------
CSEG	SEGMENT PARA PUBLIC 'CODE'
	ASSUME	CS:CSEG, DS:NOTHING, ES:NOTHING, SS:NOTHING

		ORG	0			;2.0 driver format
;----------------------------------------------------------------------
; DEVICE HEADER
; This information is used by DOS to install and identify the driver.
;----------------------------------------------------------------------
LINK		DD	-1			;Link to next driver

;	BIT NUMBER	111111
;	OF ATTR WORD	5432109876543210

ATTR		DW	1110100000000000B	;Device attribute word

 		DW	OFFSET STRAT		;Strategy entry point
		DW	OFFSET INTR		;Interrupt entry point
		DB	"NUL2    "		;8-char device name
;----------------------------------------------------------------------
; LOCAL DATA - other data interspersed with the code.
;----------------------------------------------------------------------
REQ		DD	?			;Far ptr to rqst hdr
CURSOR		DW	?			;Current cursor position
CURPG		DB	?			;Current video page

FNMSG$		DB	OFFSET FNMSG_END-$-1	;Length of message
		DB	"NUL2: FN "
FNNUM		DB	"XX = "			;Put fn number here
FNMSG_END	EQU	$

;----------------------------------------------------------------------
; The function dispatch table. Each entry is the near offset of the
; routine to service the corresponding function.
;----------------------------------------------------------------------
JMPTBL		DW	FN0,  FNX, FNX, FN3, FN4, FN5, FN6, FN7
		DW	FN8,  FN9, FNA, FNB, FNC, FND, FNE, FNF
		DW	FN10

MAX_FN		EQU	($-OFFSET JMPTBL) / 2 - 1

;======================================================================
; STRATEGY ROUTINE
; A pointer to the request header is passed in ES:BX and is saved
; locally. The caller then immediately calls the interrupt routine.
;----------------------------------------------------------------------
STRAT		PROC	FAR
	ASSUME	CS:CSEG, DS:NOTHING, ES:NOTHING, SS:NOTHING

		MOV	WORD PTR CS:[REQ][0],BX	;Save offset
		MOV	WORD PTR CS:[REQ][2],ES	; and segment
		RET

STRAT		ENDP

;======================================================================
; INTERRUPT ROUTINE
; The interrupt routine is called to interpret the request header and
; perform the indicated command.
;----------------------------------------------------------------------
INTR		PROC	FAR
	ASSUME	CS:CSEG, DS:NOTHING, ES:NOTHING, SS:NOTHING
;----------------------------------------------------------------------
; All registers must be saved. The interrupt flag (IF) and direction
; flag (DF) must also be preserved. The DOS stack has about 40-50 bytes
; free. If more are required, the driver should set up its own stack.
;----------------------------------------------------------------------
		PUSHF
		PUSH	AX
		PUSH	BX
		PUSH	CX
		PUSH	DX
		PUSH	SI
		PUSH	DI
		PUSH	DS
		PUSH	ES

		PUSH	CS
		POP	DS			;Set DS to this seg
	ASSUME	DS:CSEG
;----------------------------------------------------------------------
; Load the dword address of the header into ES:DI for easy access.
;----------------------------------------------------------------------
		LES	DI,[REQ]		;Point ES:DI to header
	ASSUME	ES:NOTHING
;----------------------------------------------------------------------
; If the function number exceeds the maximum function supported, don't
; go through the jump table. Otherwise, convert the function to an
; index and jump.
;----------------------------------------------------------------------
		SUB	BH,BH			;Set BH=0
		MOV	BL,ES:[DI+2]		;Get fn number in BL
		CMP	BL,MAX_FN		;If fn > max, invalid
		JA	FNX

		SHL	BX,1			 ;Change to word offset
		JMP	WORD PTR CS:[JMPTBL][BX] ;Jump to routine
;----------------------------------------------------------------------
; This routine is executed for functions that aren't explicitly
; supported. Both DONE and ERROR bits are set and the error code is set
; to unknown command.
;----------------------------------------------------------------------
MSGX$:		LENSTR	"Not Supported By Driver"
FNX:
		MOV	SI,OFFSET MSGX$		;Message tail
		CALL	DISPLAY_MSG

		OR	BYTE PTR ES:[DI+4],80H	;Set ERROR bit
		MOV	BYTE PTR ES:[DI+5],3	;Error = unknown cmd
		JMP	INTR_EXIT
;----------------------------------------------------------------------
; (3) IOCTL READ FROM DEVICE (If bit 14 set in attribute word)
; Return indicating that 0 bytes were read (EOF).
;----------------------------------------------------------------------
MSG3$:		LENSTR	"IOCTL Read From Device"
FN3:
		MOV	SI,OFFSET MSG3$		;Message tail
		CALL	DISPLAY_MSG

		MOV	WORD PTR ES:[DI+18],0	;No bytes read
		JMP	INTR_EXIT
;----------------------------------------------------------------------
; (4) READ FROM DEVICE
; Return indicating that 0 bytes were read (EOF).
;----------------------------------------------------------------------
MSG4$:		LENSTR	"Read From Device"
FN4:
		MOV	SI,OFFSET MSG4$		;Message tail
		CALL	DISPLAY_MSG

		MOV	WORD PTR ES:[DI+18],0	;No bytes read
		JMP	INTR_EXIT
;----------------------------------------------------------------------
; (5) NONDESTRUCTIVE READ FROM DEVICE
; Set BUSY bit to indicate no character available.
;----------------------------------------------------------------------
MSG5$:		LENSTR	"Nondestructive Read From Device"
FN5:
		MOV	SI,OFFSET MSG5$		;Message tail
		CALL	DISPLAY_MSG

		OR	BYTE PTR ES:[DI+4],20H	;Set BUSY bit
		JMP	INTR_EXIT
;----------------------------------------------------------------------
; (6) RETURN DEVICE INPUT STATUS
; Set BUSY bit to indicate no character available.
;----------------------------------------------------------------------
MSG6$:		LENSTR	"Return Device Input Status"
FN6:
		MOV	SI,OFFSET MSG6$		;Message tail
		CALL	DISPLAY_MSG

		OR	BYTE PTR ES:[DI+4],20H	;Set BUSY bit
		JMP	INTR_EXIT
;----------------------------------------------------------------------
; (7) FLUSH DEVICE INPUT BUFFER
; Simply say we did it.
;----------------------------------------------------------------------
MSG7$:		LENSTR	"Flush Device Input Buffer"
FN7:
		MOV	SI,OFFSET MSG7$		;Message tail
		CALL	DISPLAY_MSG

		JMP	INTR_EXIT
;----------------------------------------------------------------------
; (8) WRITE TO DEVICE
; Indicates that all characters have been accepted.
;----------------------------------------------------------------------
MSG8$:		LENSTR	"Write To Device"
FN8:
		MOV	SI,OFFSET MSG8$		;Message tail
		CALL	DISPLAY_MSG

		JMP	INTR_EXIT
;----------------------------------------------------------------------
; (9) WRITE TO DEVICE WITH VERIFY
; Indicates that all characters have been accepted.
;----------------------------------------------------------------------
MSG9$:		LENSTR	"Write To Device With Verify"
FN9:
		MOV	SI,OFFSET MSG9$		;Message tail
		CALL	DISPLAY_MSG

		JMP	INTR_EXIT
;----------------------------------------------------------------------
; (A) RETURN DEVICE OUTPUT STATUS
; Return BUSY = 0 to indicate no characters are in the output buffer.
;----------------------------------------------------------------------
MSGA$:		LENSTR	"Return Device Output Status"
FNA:
		MOV	SI,OFFSET MSGA$		;Message tail
		CALL	DISPLAY_MSG

		JMP	INTR_EXIT
;----------------------------------------------------------------------
; (B) FLUSH DEVICE OUTPUT BUFFER
; Return success.
;----------------------------------------------------------------------
MSGB$:		LENSTR	"Flush Device Output Buffer"
FNB:
		MOV	SI,OFFSET MSGB$		;Message tail
		CALL	DISPLAY_MSG

		JMP	SHORT INTR_EXIT
;----------------------------------------------------------------------
; (C) IOCTL WRITE TO DEVICE (If bit 14 set in attribute word)
; Indicates that all characters have been accepted.
;----------------------------------------------------------------------
MSGC$:		LENSTR	"IOCTL Write To Device"
FNC:
		MOV	SI,OFFSET MSGC$		;Message tail
		CALL	DISPLAY_MSG

		JMP	SHORT INTR_EXIT
;----------------------------------------------------------------------
; (D) OPEN DEVICE (If bit 11 set in attribute word)
; Return success.
;----------------------------------------------------------------------
MSGD$:		LENSTR	"Open Device"
FND:
		MOV	SI,OFFSET MSGD$		;Message tail
		CALL	DISPLAY_MSG

		JMP	SHORT INTR_EXIT
;----------------------------------------------------------------------
; (E) CLOSE DEVICE (If bit 11 set in attribute word)
; Return success.
;----------------------------------------------------------------------
MSGE$:		LENSTR	"Close Device"
FNE:
		MOV	SI,OFFSET MSGE$		;Message tail
		CALL	DISPLAY_MSG
		JMP	SHORT INTR_EXIT
;----------------------------------------------------------------------
; (F) REMOVEABLE MEDIA (If bit 11 set in attribute word)
; Busy means not removable. (Not applicable to char devices.)
;----------------------------------------------------------------------
MSGF$:		LENSTR	"Removable Media"
FNF:
		MOV	SI,OFFSET MSGF$		;Message tail
		CALL	DISPLAY_MSG

		OR	BYTE PTR ES:[DI+4],20H	;Set BUSY bit
		JMP	SHORT INTR_EXIT
;----------------------------------------------------------------------
; (10) OUTPUT UNTIL BUSY (If bit 13 set in attribute word)
; Indicate all characters accepted.
;----------------------------------------------------------------------
MSG10$:		LENSTR	"Output Until Busy"
FN10:
		MOV	SI,OFFSET MSG10$		;Message tail
		CALL	DISPLAY_MSG

		;Drop through to common exit routine
;----------------------------------------------------------------------
; All functions use this common driver exit.
;----------------------------------------------------------------------
INTR_EXIT:
		MOV	AH,2			;Set cursor position
		MOV	BH,[CURPG]		; for current page
		MOV	DX,[CURSOR]		; to previous value
		INT	10H			; thru BIOS

		OR	BYTE PTR ES:[DI+4],1	;Always set DONE bit

		POP	ES			;Restore all registers
	ASSUME	ES:NOTHING
		POP	DS
	ASSUME	DS:NOTHING
		POP	DI
		POP	SI
		POP	DX
		POP	CX
		POP	BX
		POP	AX
		POPF

		RET				;Exit to caller

INTR		ENDP


;======================================================================
; DISPLAY_MSG (Near, internal)
;
; 1. Get current video information.
; 2. Output standard message prefix.
; 3. Output the string argument.
;----------------------------------------------------------------------
; ENTRY:
;	DS:SI -> db LEN, STRING
;----------------------------------------------------------------------
; Changes: AX BX CX DX SI
;----------------------------------------------------------------------
DISPLAY_MSG	PROC	NEAR
	ASSUME	CS:CSEG, DS:CSEG, ES:NOTHING, SS:NOTHING

		PUSH	SI			;Save message tail
;----------------------------------------------------------------------
; Get information about the current video mode. Don't print if not in a
; text mode.
;----------------------------------------------------------------------
		MOV	AH,0FH			;Get current video info
		INT	10H			; thru BIOS
		MOV	[CURPG],BH		;Returns page in BH
						;Returns cols in AH
		CMP	AL,3
		JBE	DM_0
		CMP	AL,7
		JE	DM_0

		POP	SI			;Discard argument
		JMP	SHORT DM_EXIT
DM_0:
;----------------------------------------------------------------------
; Clear the top line of the display and set the attribute using the
; scroll active page down video function.
; The previous BIOS call returns the number of columns (1-based) in AH.
; This is adjusted to a 0-based column number in DL.
;----------------------------------------------------------------------
		MOV	DL,AH			;Number of columns
		SUB	DH,DH			;Clear DH
		DEC	DL			; to column number
		MOV	AX,701H			;Blank one line
		MOV	BH,70H			;Reverse video
		SUB	CX,CX			;Upper right corner
		INT	10H			; thru BIOS
;----------------------------------------------------------------------
; Convert the function number to two ascii characters using the
; undocumented opcode variant of AAM to divide AL by 10h; dividend goes
; in AH, remainder in AL.
;----------------------------------------------------------------------
		MOV	AL,ES:[DI+2]		;Get fn number in AL
		;AAM 10h
		DB	0D4H, 10H		;Unpack hex digits
		OR	AX,3030H		;Convert to ascii

		CMP	AL,"9"
		JBE	DM_1A
		ADD	AL,7
DM_1A:
		CMP	AH,"9"
		JBE	DM_1B
		ADD	AH,7
DM_1B:
		XCHG	AH,AL			;Put in correct order
		MOV	WORD PTR [FNNUM],AX	;Store in message
;----------------------------------------------------------------------
; Display the function number on the display.
;----------------------------------------------------------------------
		MOV	AH,3			;Get cursor position
		MOV	BH,[CURPG]		; for current page
		INT	10H			; thru BIOS
		MOV	[CURSOR],DX		;DH=row, DL=col

		MOV	AH,2			;Set cursor position
		MOV	BH,[CURPG]		; for current page
		SUB	DX,DX			; to row 0, col 0
		INT	10H			; thru BIOS
;----------------------------------------------------------------------
; Output the common string prefix and function number.
;----------------------------------------------------------------------
		CLD				;String moves forward

		MOV	SI,OFFSET FNMSG$
		CALL	DM_2A
;----------------------------------------------------------------------
; Fall thru and output message tail.
;----------------------------------------------------------------------
		POP	SI			;Retrieve message tail
DM_2A:
		SUB	CH,CH
		LODSB
		MOV	CL,AL
DM_2B:
		MOV	AH,0EH			;Write TTY
		LODSB				;Char in AL
		INT	10H			; thru BIOS
		LOOP	DM_2B
;----------------------------------------------------------------------
; This RET returns both to the local call above and to the main call.
;----------------------------------------------------------------------
DM_EXIT:
		RET

DISPLAY_MSG	ENDP

LAST_BYTE	EQU	$

;======================================================================
; FN0 - Initialize Driver
; This function is called only once: when the driver is loaded.
; It performs the init and displays a message before returning to DOS.
;----------------------------------------------------------------------
INSTALL$	DB	13,10,10,"NUL2 Sample Character Device Driver"
		DB	13,10,"Copyright (c) 1992, Robert L. Hummel"
		DB	13,10,"From The DOS Authority",13,10,10,"$"

MSG0$:		LENSTR	"Installation"

FN0		PROC	NEAR
	ASSUME	CS:CSEG, DS:CSEG, ES:NOTHING, SS:NOTHING
;----------------------------------------------------------------------
; Display our intialization message.
;----------------------------------------------------------------------
		MOV	AH,9			;Display string
		MOV	DX,OFFSET INSTALL$	;String offset
		INT	21H			; thru DOS
;----------------------------------------------------------------------
; Write the message bar.
;----------------------------------------------------------------------
		MOV	SI,OFFSET MSG0$
		CALL	DISPLAY_MSG
;----------------------------------------------------------------------
; Return a pointer to the end of our code which is to remain in memory.
;----------------------------------------------------------------------
		MOV	WORD PTR ES:[DI+14],OFFSET LAST_BYTE
		MOV	WORD PTR ES:[DI+16],CS

		JMP	INTR_EXIT		;Use common exit

FN0		ENDP

CSEG		ENDS
		END
